import json
import re
import subprocess
import sys
import tempfile
import textwrap
from pathlib import Path
from typing import *  # type: ignore

import pytest

import rio.docs

CODE_BLOCK_PATTERN = re.compile(r"```(.*?)```", re.DOTALL)


all_documented_objects = list(rio.docs.find_documented_objects())
all_documented_objects.sort(key=lambda obj: obj.__name__)


def ruff(*args: str | Path) -> subprocess.CompletedProcess:
    return subprocess.run(
        [sys.executable, "-m", "ruff", *map(str, args)],
        check=True,
        capture_output=True,
        text=True,
    )


def get_code_blocks(obj: type | Callable) -> list[str]:
    """
    Returns a list of all code blocks in the docstring of a component.
    """
    docstring = obj.__doc__

    # No docs?
    if not docstring:
        return []

    docstring = textwrap.dedent(docstring)

    # Find any contained code blocks
    result: list[str] = []
    for match in CODE_BLOCK_PATTERN.finditer(docstring):
        block: str = match.group(1)

        # Split into language and source
        linebreak = block.find("\n")
        assert linebreak != -1
        language = block[:linebreak]
        block = block[linebreak + 1 :]

        # Make sure a language is specified
        assert language, "The code block has no language specified"

        result.append(block)

    return result


def ruff_format(source_code: str) -> str:
    # Write the source code to a temporary file
    temp_file_path = Path(tempfile.gettempdir()) / "rio test suite tempfile.py"

    temp_file_path.write_text(source_code, encoding="utf8")

    # Run ruff to format the source code in the temporary file
    ruff("format", temp_file_path)

    # Read the formatted source code
    return temp_file_path.read_text(encoding="utf8")


def ruff_check(source_code: str) -> list[str]:
    """
    Checks the given source code using `ruff`. Returns any encountered problems.
    """
    # Dump the source to a file, and implicitly define/import some stuff
    temp_file_path = Path(tempfile.gettempdir()) / "rio test suite tempfile.py"

    temp_file_path.write_text(
        f"""
import pathlib
import rio

# Importing `Path` directly causes ruff to complain about a redefinition
Path = pathlib.Path
self = rio.Spacer()

{source_code}
""",
        encoding="utf8",
    )

    # Run ruff to format the source code in the temporary file
    proc = ruff(
        "check",
        temp_file_path,
        "--ignore=E402",  # Caused by the injected imports
        "--output-format=json",
    )

    output = json.loads(proc.stdout)
    assert isinstance(output, list), output

    # Parse the output
    result: list[str] = []

    for entry in output:
        result.append(entry["message"])

    return result


@pytest.mark.parametrize("obj", all_documented_objects)
def test_code_block_is_formatted(obj: type | Callable) -> None:
    # Make sure all code blocks are formatted according to ruff
    for source in get_code_blocks(obj):
        formatted_source = ruff_format(source)

        # Ruff often inserts 2 empty lines between definitions, but that's
        # really not necessary in docstrings. Collapse them to a single empty
        # line.
        source = source.replace("\n\n\n", "\n\n")
        formatted_source = formatted_source.replace("\n\n\n", "\n\n")

        assert formatted_source == source


@pytest.mark.parametrize("obj", all_documented_objects)
def test_analyze_code_block(obj: type | Callable) -> None:
    # A lot of snippets are missing context, so it's only natural that ruff will
    # find issues with the code. There isn't really anything we can do about it,
    # so we'll just skip those object.
    if obj in (rio.App, rio.Color, rio.UserSettings):
        pytest.xfail()

    # Make sure ruff is happy with all code blocks
    for source in get_code_blocks(obj):
        errors = ruff_check(source)
        assert not errors, errors



from pathlib import Path

from . import icon_registry

__all__ = ["register_icon_set", "register_icon"]


def register_icon_set(
    icon_set_name: str,
    icon_set_archive_path: Path,
) -> None:
    """
    Add an icon set to the global registry. This allows the icons to be accessed
    as `set-name/icon-name` or `set-name/icon-name:variant`.

    There must not already be a set with the given name.

    The icon set is a `.tar.xz` compressed archive and must contain exactly one
    directory, which must be named identically to the icon set. Files located in
    the root of that directory can be accessed as `set-name/icon-name`. Files
    located in a subdirectory can be accessed as `set-name/icon-name:variant`.

    For SVG files to work as icons...

    - They must have a `viewBox` attribute, but no height or width
    - They must contain exactly one XML root node: `<svg>...</svg>`. This also
      goes for comments!
    - Rio colors paths by assigning a `fill` to the SVG root. This only works as
      long as the SVG paths don't have a `<style>` assigned already.


    ## Parameters

    `icon_set_name`: The name of the new icon set. This will be used to access
        the icons.

    `icon_set_archive_path`: The path to the `.tar.xz` archive containing the
        icon set.
    """
    if icon_set_name in icon_registry.icon_set_archives:
        raise ValueError(
            f"There is already an icon set named `{icon_set_name}`"
        )

    icon_registry.icon_set_archives[icon_set_name] = icon_set_archive_path


def register_icon(
    icon_source: Path,
    icon_set_name: str,
    icon_name: str,
    variant_name: str | None = None,
) -> None:
    """
    Add a single icon to the global registry. This allows the icon to be
    accessed as `set-name/icon-name` or `set-name/icon-name:variant`.

    `icon_source` needs to be the path to a single SVG file. For SVG files
    to work as icons...

    - They must have a `viewBox` attribute, but no height or width
    - They must contain exactly one XML root node: `<svg>...</svg>`. This
      also goes for comments!
    - Rio colors paths by assigning a `fill` to the SVG root. This only
      works as long as the SVG paths don't have a `<style>` assigned
      already.


    ## Parameters

    `icon_source`: The path to the SVG file containing the icon.

    `set_name`: The name of the new icon set. This will be used to access
        the icons.

    `icon_name`: The name of the icon. This will be used to access the
        icon.

    `variant_name`: The name of the variant. This will be used to access
        the icon. If not specified, the default variant will be used.
    """

    # Try to load the icon
    svg_source = icon_source.read_text(encoding="utf8")

    # Add it to the icon registry's cache
    if variant_name is None:
        name = f"{icon_set_name}/{icon_name}"
    else:
        name = f"{icon_set_name}/{icon_name}:{variant_name}"

    icon_registry.cached_icons[name] = svg_source



from typing import *  # type:ignore

# <additional-imports>
import pandas as pd
import plotly.express as px

import rio

# </additional-imports>


# <component>
class CryptoCard(rio.Component):
    """
    Handle and display cryptocurrency-related data.

    Show a card with a line plot of the last 50 data points of the
    cryptocurrency, the cryptocurrency's logo, the name and ticker symbol of the
    cryptocurrency, the amount of the cryptocurrency, and the amount of the
    cryptocurrency in USD.


    ## Attributes

    `data`: Historical data of the fetched crypto coins.

    `coin`: Name of the selected coin.

    `coin_amount`: Representing the amount of the selected coin.

    `coin_ticker`: Representing the ticker symbol of the cryptocurrency.

    `logo_url`: Representing the URL of the cryptocurrency's logo.
    """

    data: pd.DataFrame
    coin: str
    coin_amount: float
    coin_ticker: str
    color: str
    logo_url: str

    def build(self) -> rio.Component:
        """
        Create a card with a line plot of the last 50 data points of the
        cryptocurrency, the cryptocurrency's logo, the name and ticker symbol of
        the cryptocurrency, the amount of the cryptocurrency, and the amount of
        the cryptocurrency in USD.

        See the approx. layout below:
        ```
        ╔══════════════════════════════ Card ═════════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━ Grid ━━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ ┏━━━━━━ Image ━━━━━┓ ┏━━━━━━━━━━━━━━ Plot ━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ e.g. BTC Logo    ┃ ┃ e.g. line plot of BTC data         ┃ ┃ ║
        ║ ┃ ┃                  ┃ ┃                                    ┃ ┃ ║
        ║ ┃ ┃                  ┃ ┃                                    ┃ ┃ ║
        ║ ┃ ┃                  ┃ ┃                                    ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃ ┏━━━━━━ Text ━━━━━━┓ ┏━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ Coin Ticker      ┃ ┃ Coin Amount                        ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃ ┏━━━━━━ Text ━━━━━━┓ ┏━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ Coin Name        ┃ ┃ Coin Amount in USD                 ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━┛ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚═════════════════════════════════════════════════════════════════╝
        ```
        """
        # create a line plot of the last 50 data points of the selected coin
        fig = px.line(
            self.data[self.coin].iloc[-50:],
            color_discrete_sequence=[self.color],
            height=200,
            width=200,
        )

        # hide and lock down axes
        fig.update_xaxes(visible=False, fixedrange=True)
        fig.update_yaxes(visible=False, fixedrange=True)

        # remove facet/subplot labels
        fig.update_layout(annotations=[], overwrite=True)

        # strip down the rest of the plot
        fig.update_layout(
            showlegend=False,
            margin=dict(t=10, l=10, b=10, r=10),
        )

        # create a grid layout for the card
        # The grid will have 4 rows and 3 columns
        # Because the width of the plot is bigger, we get a ratio of 2:1
        grid = rio.Grid(
            column_spacing=0.5,
            row_spacing=1,
            margin=2,
        )

        # Image with grid height 2
        grid.add(
            rio.Image(
                rio.URL(
                    self.logo_url
                ),  # logo_url = e.g. "https://cryptologos.cc/logos/bitcoin-btc-logo.svg?v=029"
                height=2,
                width=2,
                align_y=0.5,
            ),
            row=0,
            column=0,
        )

        # Plot with grid height 2
        grid.add(
            rio.Plot(
                figure=fig,
                corner_radius=0,
                height=4,
                background=self.session.theme.neutral_color,
            ),
            row=0,
            column=1,
            height=2,
        )

        # Text with coin name and grid height 1
        grid.add(
            rio.Text(
                self.coin.capitalize(),
                style=rio.TextStyle(font_size=1.2, font_weight="bold"),
                align_x=0,
            ),
            row=2,
            column=0,
        )

        # Text with coin amount and grid height 1
        grid.add(
            rio.Text(
                f"{self.coin_amount:.6f} {self.coin_ticker}",
                align_x=0,
                style=rio.TextStyle(font_size=1.2, font_weight="bold"),
            ),
            row=2,
            column=1,
        )

        # Text with coin ticker and grid height 1
        grid.add(
            rio.Row(
                rio.Text(self.coin_ticker),
                rio.Text(" / USD", style="dim"),
                align_x=0,
            ),
            row=3,
            column=0,
        )
        # Text with coin amount in USD and grid height 1
        usd_amount = self.coin_amount * self.data[self.coin].iloc[-1]
        grid.add(rio.Text(f"{usd_amount:,.2f} USD", align_x=0), row=3, column=1)

        return rio.Card(
            grid,
            height=13,
        )


# </component>



import shutil
import sys
import webbrowser
from pathlib import Path

import coverage
import pytest

project_dir = Path(__file__).parent.parent
html_dir = project_dir / "htmlcov"

# Remove old files
cov_file_path = project_dir / ".coverage"
try:
    cov_file_path.unlink()
except FileNotFoundError:
    pass
try:
    shutil.rmtree(html_dir)
except FileNotFoundError:
    pass

# Run unit tests with coverage
cov = coverage.Coverage(branch=True, source=["rio"])
cov.start()

pytest.main(["tests"])

cov.stop()
cov.save()

# Generate the HTML report
cov.html_report()

if "--no-open" not in sys.argv:
    html_path = html_dir / "index.html"
    webbrowser.open(html_path.as_uri())



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import *  # type: ignore

from uniserde import JsonDoc

from .. import maybes
from .fundamental_component import FundamentalComponent

if TYPE_CHECKING:
    import numpy  # type: ignore
    import pandas  # type: ignore
    import polars  # type: ignore


__all__ = ["Table"]


TableValue = int | float | str


@final
class Table(FundamentalComponent):
    """
    A table of data.

    Tables are a way to display data in a grid, with rows and columns. They are
    very useful for displaying data that is naturally tabular, such as
    spreadsheets, databases, or CSV files. Tables can be sorted by clicking on
    the column headers.


    ## Attributes

    `data`: The data to display.

    `show_row_numbers`: Whether to show row numbers on the left side of the table.


    ## Examples

    A simple table with some data:

    ```python
    rio.Table(
        data={
            "Name": ["Alice", "Bob", "Charlie"],
            "Age": [25, 30, 35],
            "City": ["New York", "San Francisco", "Los Angeles"],
        }
    )
    ```


    ## Metadata

    `experimental`: True
    """

    data: (
        pandas.DataFrame
        | polars.DataFrame
        | Mapping[str, Iterable[TableValue]]
        | Iterable[Iterable[TableValue]]
        | numpy.ndarray
    )
    show_row_numbers: bool = True

    def _custom_serialize(self) -> JsonDoc:
        return {
            "data": self._data_to_json(),  # type: ignore (variance)
        }

    def _data_to_json(
        self,
    ) -> dict[str, list[TableValue]] | list[list[TableValue]]:
        data = self.data

        if isinstance(data, maybes.PANDAS_DATAFRAME_TYPES):
            return data.to_dict(orient="list")  # type: ignore

        if isinstance(data, maybes.POLARS_DATAFRAME_TYPES):
            return data.to_dict(as_series=False)  # type: ignore

        if isinstance(data, Mapping):
            return dict(data)  # type: ignore (wtf?)

        if isinstance(data, maybes.NUMPY_ARRAY_TYPES):
            return data.tolist()  # type: ignore

        data = cast(Iterable[Iterable[TableValue]], data)
        return [row if isinstance(row, list) else list(row) for row in data]


Table._unique_id = "Table-builtin"



from __future__ import annotations

from collections.abc import Mapping, Sequence
from dataclasses import KW_ONLY, dataclass
from typing import Any, Generic, Literal, TypeVar, final

from uniserde import JsonDoc

import rio.docs

from .fundamental_component import FundamentalComponent

__all__ = [
    "Dropdown",
    "DropdownChangeEvent",
]

T = TypeVar("T")


@final
@rio.docs.mark_constructor_as_private
@dataclass
class DropdownChangeEvent(Generic[T]):
    """
    Holds information regarding a dropdown change event.

    This is a simple dataclass that stores useful information for when the user
    selects an option in a `Dropdown`. You'll typically receive this as argument
    in `on_change` events.

    ## Attributes

    `value`: The new `selected_value` of the `Dropdown`.
    """

    value: T


@final
class Dropdown(FundamentalComponent, Generic[T]):
    """
    A dropdown menu for selecting from one of several options.

    Dropdowns present the user with a list of options, allowing them to select
    exactly one. In their default state dropdowns are compact and display the
    currently selected option. When activated, a popup menu appears with a list
    of all options. Typing while the dropdown is opened will filter the options.

    ## Attributes

    `options`: A mapping from option names to values. The names are displayed
        in the dropdown menu, and the corresponding value is returned when
        the user selects the option. The values must be comparable.

    `label`: A short text to display next to the dropdown.

    `selected_value`: The value of the currently selected option.

    `is_sensitive`: Whether the dropdown should respond to user input.

    `is_valid`: Visually displays to the user whether the current option is
        valid. You can use this to signal to the user that their input needs
        to be changed.

    `on_change`: Triggered whenever the user selects an option.


    ## Examples

    Here's a simple example that allows the user to select an option and
    displays it back to them:

    ```python
    class MyComponent(rio.Component):
        selected_value: str = "a"

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Dropdown(
                    options={
                        "Option A": "a",
                        "Option B": "b",
                        "Option C": "c",
                    },
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the text.
                    selected_value=self.bind().selected_value,
                    label="Enter a Text",
                ),
                rio.Text(f"You've chosen {self.selected_value}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    selects an option:

    ```python
    class MyComponent(rio.Component):
        selected_value: str = "a"

        def on_value_change(self, event: rio.DropdownChangeEvent):
            # This function will be called whenever the user selects an. We'll
            # display the new selection in addition to updating our own
            # attribute.
            self.selected_value = event.value
            print(f"You've chosen {self.selected_value}")

        def build(self) -> rio.Component:
            return rio.Dropdown(
                options={
                    "Option A": "a",
                    "Option B": "b",
                    "Option C": "c",
                },
                selected_value=self.selected_value,
                label="Enter a Text",
                on_change=self.on_value_change,
            )
    ```
    """

    options: Mapping[str, T]
    _: KW_ONLY
    label: str
    selected_value: T
    is_sensitive: bool
    is_valid: bool
    on_change: rio.EventHandler[DropdownChangeEvent[T]]

    def __init__(
        self,
        options: Mapping[str, T] | Sequence[T],
        *,
        label: str = "",
        selected_value: T | None = None,
        on_change: rio.EventHandler[DropdownChangeEvent[T]] = None,
        is_sensitive: bool = True,
        is_valid: bool = True,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        if not options:
            raise ValueError("`Dropdown` must have at least one option.")

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        if isinstance(options, Sequence):
            options = {str(value): value for value in options}

        self.options = options
        self.label = label
        self.on_change = on_change
        self.is_sensitive = is_sensitive
        self.is_valid = is_valid

        # This is an unsafe assignment, because the value could be `None`. This
        # will be fixed in `__post_init__`, once the attribute bindings have been
        # initialized.
        self.selected_value = selected_value  # type: ignore

    def __post_init__(self) -> None:
        # Make sure a value is selected
        if self.selected_value is None:
            self.selected_value = next(iter(self.options.values()))

    def _fetch_selected_name(self) -> str:
        # The frontend works with names, not values. Get the corresponding
        # name.

        # Avoid hammering a potential attribute binding
        selected_value = self.selected_value

        # Fetch the name
        for name, value in self.options.items():
            if value == selected_value:
                return name
        else:
            # If nothing matches, just select the first option
            return next(iter(self.options.keys()))

    def _custom_serialize(self) -> JsonDoc:
        result: JsonDoc = {
            "optionNames": list(self.options.keys()),
            "selectedName": self._fetch_selected_name(),
        }

        return result

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        # The frontend works with names, not values. Get the corresponding
        # value.
        try:
            selected_value = self.options[msg["name"]]
        except KeyError:
            # Invalid names may be sent due to lag between the frontend and
            # backend. Ignore them.
            return

        self._apply_delta_state_from_frontend(
            {"selected_value": selected_value}
        )

        # Trigger the event
        await self.call_event_handler(
            self.on_change, DropdownChangeEvent(self.selected_value)
        )

        # Refresh the session
        await self.session._refresh()


Dropdown._unique_id = "Dropdown-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from datetime import date
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio.docs

from .fundamental_component import FundamentalComponent

__all__ = [
    "DateChangeEvent",
    "Calendar",
]


@final
@rio.docs.mark_constructor_as_private
@dataclass
class DateChangeEvent:
    """
    Holds information regarding a date change event.

    This is a simple dataclass that stores useful information for when the user
    changes the date in a `Calendar` or `DateInput`. You'll typically receive
    this as argument in `on_change` events.

    ## Attributes

    `value`: The newly selected date.
    """

    value: date


class Calendar(FundamentalComponent):
    """
    Allows the user to pick a date from a calendar.

    Calendars are large components, that display one month at a time. The user
    can switch between months and years, and select a day from the displayed
    month.

    For a similar, but more compact component consider using `rio.DateInput`.


    ## Attributes

    `value`: The currently selected date.

    `on_change`: Triggered whenever the user selects a new date.


    ## Examples

    Here's a simple example that allows the user to select a data and displays
    it back to them:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Calendar(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the value.
                    value=self.bind().value,
                ),
                rio.Text(f"You've selected: {self.value}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    picks a new date:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def on_value_change(self, event: rio.DateChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.value = event.value
            print(f"You've selected: {self.value}")

        def build(self) -> rio.Component:
            return rio.Calendar(
                value=self.value,
                on_change=self.on_value_change,
            )
    ```


    ## Metadata

    `experimental`: True
    """

    value: date

    _: KW_ONLY

    on_change: rio.EventHandler[DateChangeEvent] = None

    def _custom_serialize(self) -> JsonDoc:
        return {
            "selectedYear": self.value.year,
            "selectedMonth": self.value.month,
            "selectedDay": self.value.day,
            "monthNamesLong": self.session._month_names_long,
            "dayNamesLong": self.session._day_names_long,
            "firstDayOfWeek": self.session._first_day_of_week,
        }

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        try:
            new_year = msg["year"]
            new_month = msg["month"]
            new_day = msg["day"]
        except KeyError:
            raise AssertionError(
                f"Frontend has sent a message with missing keys: {msg}"
            )

        # Build the new date, still looking out for invalid values
        try:
            self.value = date(new_year, new_month, new_day)
        except ValueError as e:
            raise AssertionError(
                f"The frontend has sent an invalid date: {new_year!r}-{new_month!r}-{new_day!r}"
            ) from e

        # Trigger the press event
        await self.call_event_handler(
            self.on_change,
            DateChangeEvent(self.value),
        )

        # Refresh the session
        await self.session._refresh()


Calendar._unique_id = "Calendar-builtin"



from __future__ import annotations

from collections.abc import Callable
from typing import *  # type: ignore

from .. import utils
from .component import Component
from .fundamental_component import FundamentalComponent
from .scroll_container import ScrollContainer

__all__ = ["HighLevelRootComponent"]


# The HighLevelRootComponent is the top-most element. It's a high-level element
# that is never re-built, which is convenient in many ways:
# 1. Every component except for the root component itself has a valid builder
# 2. The JS code is simpler because the root component can't have an
#    alignment or margin
# 3. Children of non-fundamental components are automatically initialized
#    correctly, so we don't need to duplicate that logic here
class HighLevelRootComponent(Component):
    """
    ## Metadata

    public: False
    """

    build_function: Callable[[], Component]
    build_connection_lost_message_function: Callable[[], Component]

    def build(self) -> Component:
        # Spawn the dev tools if running in debug mode.
        #
        # The browser handles scrolling automatically if the content grows too
        # large for the window, but when the dev tools are visible, they would
        # appear between the scroll bar and the scrolling content. To prevent
        # this, we'll wrap the user content in a ScrollContainer.
        #
        # Conditionally inserting this ScrollContainer would make a bunch of
        # code more messy, so we'll *always* insert the ScrollContainer but
        # conditionally disable it with `scroll='never'`.
        if self.session._app_server.debug_mode:
            # Avoid a circular import
            import rio.debug.dev_tools

            dev_tools = rio.debug.dev_tools.DevToolsSidebar()

            scroll = "auto"
        else:
            dev_tools = None
            scroll = "never"

        # Build the user's root component
        user_root = utils.safe_build(self.build_function)

        # Wrap it up in a scroll container, so the dev-tools don't scroll
        user_content = ScrollContainer(
            user_root,
            scroll_x=scroll,
            scroll_y=scroll,
        )

        return FundamentalRootComponent(
            user_content,
            utils.safe_build(self.build_connection_lost_message_function),
            dev_tools=dev_tools,
        )


class FundamentalRootComponent(FundamentalComponent):
    """
    ## Metadata

    public: False
    """

    content: Component
    connection_lost_component: Component
    dev_tools: Component | None


FundamentalRootComponent._unique_id = "FundamentalRootComponent-builtin"



import abc
import asyncio

from uniserde import JsonDoc

__all__ = [
    "AbstractTransport",
    "TransportClosedIntentionally",
    "TransportInterrupted",
]


class AbstractTransport(abc.ABC):
    """
    Represents a communication channel between a `Session` and the client.
    """

    def __init__(self) -> None:
        self.closed = asyncio.Event()

    @abc.abstractmethod
    async def send(self, message: str, /) -> None:
        """
        Send the message if possible. If the transport is closed, do nothing.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def receive(self) -> JsonDoc:
        """
        Return the next message received from the client. If the transport is
        closed, throw either `TransportInterrupted` or
        `TransportClosedIntentionally`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """
        Close the connection. Set `self.closed` when finished.
        """
        raise NotImplementedError


class TransportClosedIntentionally(Exception):
    pass


class TransportInterrupted(Exception):
    pass



from __future__ import annotations

from typing import *  # type: ignore

import rio


# <component>
class Footer(rio.Component):
    """
    A simple, static component which displays a footer with the company name and
    website name.
    """

    def build(self) -> rio.Component:
        return rio.Card(
            content=rio.Column(
                rio.Icon("rio/logo:fill", width=5, height=5),
                rio.Text("Buzzwordz Inc.", justify="center"),
                rio.Text(
                    "Hyper Dyper Website",
                    style="dim",
                ),
                spacing=1,
                margin=2,
                align_x=0.5,
            ),
            color="hud",
            corner_radius=0,
        )


# </component>



from __future__ import annotations

from typing import Literal

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "CodeExplorer",
]


class CodeExplorer(FundamentalComponent):
    """
    ## Metadata

    public: False
    """

    source_code: str
    build_result: rio.Component
    line_indices_to_component_keys: list[str | None]

    style: Literal["horizontal", "vertical"] = "horizontal"


CodeExplorer._unique_id = "CodeExplorer-builtin"



from __future__ import annotations

import copy
import io
from typing import TYPE_CHECKING, Literal, cast, final

from uniserde import JsonDoc

from .. import fills, maybes
from .fundamental_component import FundamentalComponent

if TYPE_CHECKING:
    import matplotlib.axes  # type: ignore
    import matplotlib.figure  # type: ignore
    import plotly.graph_objects  # type: ignore


__all__ = ["Plot"]


@final
class Plot(FundamentalComponent):
    """
    Displays a `matplotlib`, `seaborn` or `plotly` plot.

    Plots are a very useful tool to visualize data. Not only that, but having
    a pretty graph in your app is a great way to make it more engaging and
    beautiful.

    Rio supports the most popular Python plotting libraries around: It can
    display plots made with `matplotlib`, `seaborn`, as well as `plotly`. Create
    a plot using the library of your choice and pass it to the `Plot` component
    to display it in your app.

    ## Attributes

    `figure`: The plot figure to display.

    `background`: The background color of the plot. If `None`, a color from
        the theme is used.

    `corner_radius`: The corner radius of the plot


    ## Examples

    Here's a minimal example using a `plotly` plot:

    ```python
    import pandas as pd
    import plotly.express as px


    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            df = pd.DataFrame(
                {
                    "x": [1, 2, 3, 4],
                    "y": [4, 3, 2, 1],
                }
            )
            fig = px.line(df, x="x", y="y", title="Sample Figure")

            return rio.Plot(fig)
    ```
    """

    figure: (
        plotly.graph_objects.Figure
        | matplotlib.figure.Figure
        | matplotlib.axes.Axes
    )
    background: fills._FillLike | None
    corner_radius: float | tuple[float, float, float, float] | None

    def __init__(
        self,
        figure: (
            plotly.graph_objects.Figure
            | matplotlib.figure.Figure
            | matplotlib.axes.Axes
        ),
        *,
        background: fills._FillLike | None = None,
        corner_radius: float | tuple[float, float, float, float] | None = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: Literal["natural", "grow"] | float = "natural",
        height: Literal["natural", "grow"] | float = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.figure = figure
        self.background = background

        if corner_radius is None:
            self.corner_radius = self.session.theme.corner_radius_small
        else:
            self.corner_radius = corner_radius

    def _custom_serialize(self) -> JsonDoc:
        # Figure
        figure = self.figure
        plot: JsonDoc

        # Plotly
        if isinstance(figure, maybes.PLOTLY_GRAPH_TYPES):
            # Make the plot transparent, so `self.background` shines through.
            figure = cast("plotly.graph_objects.Figure", copy.copy(figure))
            figure.update_layout(
                {
                    "plot_bgcolor": "rgba(0,0,0,0)",
                    "paper_bgcolor": "rgba(0,0,0,0)",
                }
            )

            plot = {
                "type": "plotly",
                "json": figure.to_json(),
            }

        # Matplotlib (+ Seaborn)
        elif isinstance(figure, maybes.MATPLOTLIB_GRAPH_TYPES):
            if isinstance(figure, maybes.MATPLOTLIB_AXES_TYPES):
                figure = figure.figure

            figure = cast("matplotlib.figure.Figure", figure)

            file = io.BytesIO()
            figure.savefig(
                file,
                format="svg",
                transparent=True,
                bbox_inches="tight",
            )

            plot = {
                "type": "matplotlib",
                "svg": bytes(file.getbuffer()).decode("utf-8"),
            }

        # Unsupported
        else:
            raise TypeError(f"Unsupported plot type: {type(figure)}")

        # Corner radius
        if isinstance(self.corner_radius, (int, float)):
            corner_radius = (self.corner_radius,) * 4
        else:
            corner_radius = self.corner_radius

        # Combine everything
        return {
            "plot": plot,
            "corner_radius": corner_radius,
        }


Plot._unique_id = "Plot-builtin"



from .chat_message import ChatMessage as ChatMessage
from .chat_suggestion_card import ChatSuggestionCard as ChatSuggestionCard
from .empty_chat_placeholder import EmptyChatPlaceholder as EmptyChatPlaceholder
from .generating_response_placeholder import (
    GeneratingResponsePlaceholder as GeneratingResponsePlaceholder,
)



corner_round_bottom_left = "styling/corner-round-bottom-left"
corner_round_bottom_right = "styling/corner-round-bottom-right"
corner_round_top_left = "styling/corner-round-top-left"
corner_round_top_right = "styling/corner-round-top-right"



from __future__ import annotations

import asyncio
import contextlib
import functools
import html
import io
import json
import logging
import secrets
import weakref
from datetime import timedelta
from pathlib import Path
from typing import *  # type: ignore
from xml.etree import ElementTree as ET

import crawlerdetect
import fastapi
import timer_dict
from PIL import Image
from uniserde import Jsonable, JsonDoc

import rio

from .. import (
    app,
    assets,
    byte_serving,
    data_models,
    errors,
    inspection,
    routing,
    utils,
)
from ..errors import AssetError
from ..icons import icon_registry
from ..transports import FastapiWebsocketTransport, MessageRecorderTransport
from ..utils import URL
from .abstract_app_server import AbstractAppServer

try:
    import plotly  # type: ignore (missing import)
except ImportError:
    plotly = None

__all__ = [
    "FastapiServer",
]


P = ParamSpec("P")


# Used to identify search engine crawlers (like googlebot) and serve them
# without needing a websocket connection
CRAWLER_DETECTOR = crawlerdetect.CrawlerDetect()


@functools.lru_cache(maxsize=None)
def _build_sitemap(base_url: rio.URL, app: rio.App) -> str:
    # Find all pages to add
    page_urls = {
        rio.URL(""),
    }

    def worker(
        parent_url: rio.URL,
        page: rio.Page,
    ) -> None:
        cur_url = parent_url / page.page_url
        page_urls.add(cur_url)

        for child in page.children:
            worker(cur_url, child)

    for page in app.pages:
        worker(rio.URL(), page)

    # Build a XML site map
    tree = ET.Element(
        "urlset",
        xmlns="http://www.sitemaps.org/schemas/sitemap/0.9",
    )

    for relative_url in page_urls:
        full_url = base_url.with_path(relative_url.path)

        url = ET.SubElement(tree, "url")
        loc = ET.SubElement(url, "loc")
        loc.text = str(full_url)

    # Done
    return ET.tostring(tree, encoding="unicode", xml_declaration=True)


@functools.lru_cache(maxsize=None)
def read_frontend_template(template_name: str) -> str:
    """
    Read a text file from the frontend directory and return its content. The
    results are cached to avoid repeated disk access.
    """
    return (utils.FRONTEND_FILES_DIR / template_name).read_text(
        encoding="utf-8"
    )


def add_cache_headers(
    func: Callable[P, Awaitable[fastapi.Response]],
) -> Callable[P, Coroutine[None, None, fastapi.Response]]:
    """
    Decorator for routes that serve static files. Ensures that the response has
    the `Cache-Control` header set appropriately.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> fastapi.Response:
        response = await func(*args, **kwargs)
        response.headers["Cache-Control"] = "max-age=31536000, immutable"
        return response

    return wrapper


class FastapiServer(fastapi.FastAPI, AbstractAppServer):
    def __init__(
        self,
        app_: app.App,
        debug_mode: bool,
        running_in_window: bool,
        internal_on_app_start: Callable[[], None] | None,
    ):
        super().__init__(
            title=app_.name,
            # summary=...,
            # description=...,
            openapi_url=None,
            # openapi_url="/openapi.json" if debug_mode else None,
            # docs_url="/docs" if debug_mode else None,
            # redoc_url="/redoc" if debug_mode else None,
            lifespan=__class__._lifespan,
        )
        AbstractAppServer.__init__(
            self,
            app_,
            running_in_window=running_in_window,
            debug_mode=debug_mode,
        )

        self.internal_on_app_start = internal_on_app_start

        # While this Event is unset, no new Sessions can be created. This is
        # used to ensure that no clients (re-)connect while `rio run` is
        # reloading the app.
        self._can_create_new_sessions = asyncio.Event()
        self._can_create_new_sessions.set()

        # Initialized lazily, when the favicon is first requested.
        self._icon_as_png_blob: bytes | None = None

        # The session tokens and Request object for all clients that have made a
        # HTTP request, but haven't yet established a websocket connection. Once
        # the websocket connection is created, these will be turned into
        # Sessions.
        self._latent_session_tokens: dict[str, fastapi.Request] = {}

        # The session tokens for all active sessions. These allow clients to
        # identify themselves, for example to reconnect in case of a lost
        # connection.
        self._active_session_tokens: dict[str, rio.Session] = {}
        self._active_tokens_by_session: dict[rio.Session, str] = {}

        # All assets that have been registered with this session. They are held
        # weakly, meaning the session will host assets for as long as their
        # corresponding Python objects are alive.
        #
        # Assets registered here are hosted under `/asset/temp-{asset_id}`. In
        # addition the server also permanently hosts other "well known" assets
        # (such as javascript dependencies) which are available under public
        # URLS at `/asset/{some-name}`.
        self._assets: weakref.WeakValueDictionary[str, assets.Asset] = (
            weakref.WeakValueDictionary()
        )

        # All pending file uploads. These are stored in memory for a limited
        # time. When a file is uploaded the corresponding future is set.
        self._pending_file_uploads: timer_dict.TimerDict[
            str, asyncio.Future[list[utils.FileInfo]]
        ] = timer_dict.TimerDict(default_duration=timedelta(minutes=15))

        # FastAPI
        self.add_api_route("/robots.txt", self._serve_robots, methods=["GET"])
        self.add_api_route("/rio/sitemap", self._serve_sitemap, methods=["GET"])
        self.add_api_route(
            "/rio/favicon.png", self._serve_favicon, methods=["GET"]
        )
        self.add_api_route(
            "/rio/frontend/assets/{asset_id:path}",
            self._serve_frontend_asset,
        )
        self.add_api_route(
            "/rio/assets/special/{asset_id:path}",
            self._serve_special_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/assets/hosted/{asset_id:path}",
            self._serve_hosted_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/assets/user/{asset_id:path}",
            self._serve_user_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/assets/temp/{asset_id:path}",
            self._serve_temp_asset,
            methods=["GET"],
        )
        self.add_api_route(
            "/rio/icon/{icon_name:path}", self._serve_icon, methods=["GET"]
        )
        self.add_api_route(
            "/rio/upload/{upload_token}",
            self._serve_file_upload,
            methods=["PUT"],
        )
        self.add_api_websocket_route("/rio/ws", self._serve_websocket)

        # This route is only used in `debug_mode`. When the websocket connection
        # is interrupted, the frontend polls this route and then either
        # reconnects or reloads depending on whether its session token is still
        # valid (i.e. depending on whether the server was restarted or not)
        self.add_api_route(
            "/rio/validate-token/{session_token}",
            self._serve_token_validation,
        )

        # The route that serves the index.html will be registered later, so that
        # it has a lower priority than user-created routes.

    async def __call__(self, scope, receive, send):
        # Because this is a single page application, all other routes should
        # serve the index page. The session will determine which components
        # should be shown.
        self.add_api_route(
            "/{initial_route_str:path}", self._serve_index, methods=["GET"]
        )

        return await super().__call__(scope, receive, send)

    @contextlib.asynccontextmanager
    async def _lifespan(self):
        await self._on_start()

        # Trigger any internal startup event
        if self.internal_on_app_start is not None:
            self.internal_on_app_start()

        try:
            yield
        finally:
            await self._on_close()

    def url_for_user_asset(self, relative_asset_path: Path) -> rio.URL:
        return rio.URL(f"/rio/assets/user/{relative_asset_path}")

    def weakly_host_asset(self, asset: assets.HostedAsset) -> str:
        """
        Register an asset with this server. The asset will be held weakly,
        meaning the server will host assets for as long as their corresponding
        Python objects are alive.

        If another asset with the same id is already hosted, it will be
        replaced.
        """
        self._assets[asset.secret_id] = asset
        return f"/rio/assets/temp/{asset.secret_id}"

    def _get_all_meta_tags(self, title: str | None = None) -> list[str]:
        """
        Returns all `<meta>` tags that should be added to the app's HTML page.
        This includes auto-generated ones, as well as those stored directly in
        the app.
        """
        if title is None:
            title = self.app.name

        # Prepare the default tags
        all_tags = {
            "og:title": title,
            "description": self.app.description,
            "og:description": self.app.description,
            "keywords": "python, web, app, rio",
            "image": "/rio/favicon.png",
            "viewport": "width=device-width, initial-scale=1",
        }

        # Add the user-defined meta tags, overwriting any automatically
        # generated ones
        all_tags.update(self.app._custom_meta_tags)

        # Convert everything to HTML
        result: list[str] = []

        for key, value in all_tags.items():
            key = html.escape(key)
            value = html.escape(value)
            result.append(f'<meta name="{key}" content="{value}">')

        return result

    async def _serve_index(
        self,
        request: fastapi.Request,
        initial_route_str: str,
    ) -> fastapi.responses.Response:
        """
        Handler for serving the index HTML page via fastapi.
        """

        # Because Rio apps are single-page, this route serves as the fallback.
        # In addition to legitimate requests for HTML pages, it will also catch
        # a bunch of invalid requests to other resources. To highlight this,
        # throw a 404 if HTML is not explicitly requested.
        #
        # Currently inactive, because this caused issues behind dumb proxies
        # that don't pass on the `accept` header.

        # if not request.headers.get("accept", "").startswith("text/html"):
        #     raise fastapi.HTTPException(
        #         status_code=fastapi.status.HTTP_404_NOT_FOUND,
        #     )

        initial_messages = list[JsonDoc]()

        is_crawler = CRAWLER_DETECTOR.isCrawler(
            request.headers.get("User-Agent")
        )
        if is_crawler:
            # If it's a crawler, immediately create a Session for it. Instead of
            # a websocket connection, outgoing messages are simply appended to a
            # list which will be included in the HTML.
            transport = MessageRecorderTransport()
            initial_messages = transport.sent_messages

            assert request.client is not None, "How can this happen?!"

            requested_url = rio.URL(str(request.url))
            try:
                session = await self.create_session(
                    initial_message=data_models.InitialClientMessage.from_defaults(
                        url=str(requested_url),
                    ),
                    transport=transport,
                    client_ip=request.client.host,
                    client_port=request.client.port,
                    http_headers=request.headers,
                )
            except routing.NavigationFailed:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail=f"Navigation to initial page `{request.url}` has failed.",
                ) from None

            session.close()

            # If a page guard caused a redirect, tell that to the crawler in a
            # language it understands
            if session.active_page_url != requested_url:
                return fastapi.responses.RedirectResponse(
                    str(session.active_page_url)
                )

            session_token = "<crawler>"

            title = " - ".join(
                page.name for page in session.active_page_instances
            )
        else:
            # Create a session token that uniquely identifies this client
            session_token = secrets.token_urlsafe()
            self._latent_session_tokens[session_token] = request

            title = self.app.name

        # Load the template
        html_ = read_frontend_template("index.html")

        html_ = html_.replace("{session_token}", session_token)

        html_ = html_.replace(
            "'{child_attribute_names}'",
            json.dumps(
                inspection.get_child_component_containing_attribute_names_for_builtin_components()
            ),
        )

        html_ = html_.replace(
            "'{ping_pong_interval}'",
            str(self.app._ping_pong_interval.total_seconds()),
        )

        html_ = html_.replace(
            "'{debug_mode}'",
            "true" if self.debug_mode else "false",
        )

        html_ = html_.replace(
            "'{running_in_window}'",
            "true" if self.running_in_window else "false",
        )

        html_ = html_.replace(
            "'{initial_messages}'", json.dumps(initial_messages)
        )

        # Since the title is user-defined, it might contain placeholders like
        # `{debug_mode}`. So it's important that user-defined content is
        # inserted last.
        html_ = html_.replace("{title}", html.escape(title))

        # The placeholder for the metadata uses unescaped `<` and `>` characters
        # to ensure that no user-defined content can accidentally contain this
        # placeholder.
        html_ = html_.replace(
            '<meta name="{meta}" />',
            "\n".join(self._get_all_meta_tags(title)),
        )

        # Respond
        return fastapi.responses.HTMLResponse(html_)

    async def _serve_robots(
        self, request: fastapi.Request
    ) -> fastapi.responses.Response:
        """
        Handler for serving the `robots.txt` file via fastapi.
        """

        # TODO: Disallow internal API routes? Icons, assets, etc?
        request_url = URL(str(request.url))
        content = f"""
User-agent: *
Allow: /

Sitemap: {request_url.with_path("/rio/sitemap")}
        """.strip()

        return fastapi.responses.Response(
            content=content,
            media_type="text/plain",
        )

    async def _serve_sitemap(
        self, request: fastapi.Request
    ) -> fastapi.responses.Response:
        """
        Handler for serving the `sitemap.xml` file via fastapi.
        """
        return fastapi.responses.Response(
            content=_build_sitemap(rio.URL(str(request.url)), self.app),
            media_type="application/xml",
        )

    async def _serve_favicon(self) -> fastapi.responses.Response:
        """
        Handler for serving the favicon via fastapi, if one is set.
        """
        # If an icon is set, make sure a cached version exists
        if self._icon_as_png_blob is None and self.app._icon is not None:
            try:
                icon_blob, _ = await self.app._icon.try_fetch_as_blob()

                input_buffer = io.BytesIO(icon_blob)
                output_buffer = io.BytesIO()

                with Image.open(input_buffer) as image:
                    image.save(output_buffer, format="png")

            except Exception as err:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail="Could not fetch the app's icon.",
                ) from err

            self._icon_as_png_blob = output_buffer.getvalue()

        # No icon set or fetching failed
        if self._icon_as_png_blob is None:
            return fastapi.responses.Response(status_code=404)

        # There is an icon, respond
        return fastapi.responses.Response(
            content=self._icon_as_png_blob,
            media_type="image/png",
        )

    async def _serve_frontend_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        response = await self._serve_file_from_directory(
            request,
            utils.FRONTEND_ASSETS_DIR,
            asset_id + ".gz",
        )
        response.headers["content-encoding"] = "gzip"

        return response

    @add_cache_headers
    async def _serve_special_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        # The python plotly library already includes a minified version of
        # plotly.js. Rather than shipping another one, just serve the one
        # included in the library.
        if asset_id == "plotly.min.js":
            if plotly is None:
                return fastapi.responses.Response(status_code=404)

            return fastapi.responses.Response(
                content=plotly.offline.get_plotlyjs(),
                media_type="text/javascript",
            )

        return fastapi.responses.Response(status_code=404)

    async def _serve_hosted_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        return await self._serve_file_from_directory(
            request,
            utils.HOSTED_ASSETS_DIR,
            asset_id,
        )

    async def _serve_user_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        return await self._serve_file_from_directory(
            request,
            self.app.assets_dir,
            asset_id,
        )

    @add_cache_headers
    async def _serve_temp_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        # Get the asset's Python instance. The asset's id acts as a secret, so
        # no further authentication is required.
        try:
            asset = self._assets[asset_id]
        except KeyError:
            return fastapi.responses.Response(status_code=404)

        # Fetch the asset's content and respond
        if isinstance(asset, assets.BytesAsset):
            return fastapi.responses.Response(
                content=asset.data,
                media_type=asset.media_type,
            )
        elif isinstance(asset, assets.PathAsset):
            return byte_serving.range_requests_response(
                request,
                file_path=asset.path,
                media_type=asset.media_type,
            )
        else:
            assert False, f"Unable to serve asset of unknown type: {asset}"

    @add_cache_headers
    async def _serve_file_from_directory(
        self,
        request: fastapi.Request,
        directory: Path,
        asset_id: str,
    ) -> fastapi.responses.Response:
        # Construct the path to the target file
        asset_file_path = directory / asset_id

        # Make sure the path is inside the hosted assets directory
        asset_file_path = asset_file_path.absolute()
        if not asset_file_path.is_relative_to(directory):
            logging.warning(
                f'Client requested asset "{asset_id}" which is not located'
                f" inside the assets directory. Somebody might be trying to"
                f" break out of the assets directory!"
            )
            return fastapi.responses.Response(status_code=404)

        return byte_serving.range_requests_response(
            request,
            file_path=asset_file_path,
        )

    @add_cache_headers
    async def _serve_icon(self, icon_name: str) -> fastapi.responses.Response:
        """
        Allows the client to request an icon by name. This is not actually the
        mechanism used by the `Icon` component, but allows JavaScript to request
        icons.
        """
        # Get the icon's SVG
        try:
            svg_source = icon_registry.get_icon_svg(icon_name)
        except AssetError:
            return fastapi.responses.Response(status_code=404)

        # Respond
        return fastapi.responses.Response(
            content=svg_source,
            media_type="image/svg+xml",
        )

    async def _serve_file_upload(
        self,
        upload_token: str,
        file_names: list[str],
        file_types: list[str],
        file_sizes: list[str],
        # If no files are uploaded `files` isn't present in the form data at
        # all. Using a default value ensures that those requests don't fail
        # because of "missing parameters".
        #
        # Lists are mutable, make sure not to modify this value!
        file_streams: list[fastapi.UploadFile] = [],
    ):
        # Try to find the future for this token
        try:
            future = self._pending_file_uploads.pop(upload_token)
        except KeyError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_400_BAD_REQUEST,
                detail="Invalid upload token.",
            )

        # Make sure the same number of values was received for each parameter
        n_names = len(file_names)
        n_types = len(file_types)
        n_sizes = len(file_sizes)
        n_streams = len(file_streams)

        if n_names != n_types or n_names != n_sizes or n_names != n_streams:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Inconsistent number of files between the different message parts.",
            )

        # Parse the file sizes
        parsed_file_sizes: list[int] = []
        for file_size in file_sizes:
            try:
                parsed = int(file_size)
            except ValueError:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Invalid file size.",
                )

            if parsed < 0:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Invalid file size.",
                )

            parsed_file_sizes.append(parsed)

        # Complete the future
        future.set_result(
            [
                utils.FileInfo(
                    name=file_names[ii],
                    size_in_bytes=parsed_file_sizes[ii],
                    media_type=file_types[ii],
                    _contents=await file_streams[ii].read(),
                )
                for ii in range(n_names)
            ]
        )

        return fastapi.responses.Response(
            status_code=fastapi.status.HTTP_200_OK
        )

    async def file_chooser(
        self,
        session: rio.Session,
        *,
        file_extensions: Iterable[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | tuple[utils.FileInfo, ...]:
        # Create a secret id and register the file upload with the app server
        upload_id = secrets.token_urlsafe()
        future = asyncio.Future[list[utils.FileInfo]]()

        self._pending_file_uploads[upload_id] = future

        # Allow the user to specify both `jpg` and `.jpg`
        if file_extensions is not None:
            file_extensions = [
                ext if ext.startswith(".") else f".{ext}"
                for ext in file_extensions
            ]

        # Tell the frontend to upload a file
        await session._request_file_upload(
            upload_url=f"/rio/upload/{upload_id}",
            file_extensions=file_extensions,
            multiple=multiple,
        )

        # Wait for the user to upload files
        files = await future

        # Raise an exception if no files were uploaded
        if not files:
            raise errors.NoFileSelectedError()

        # Ensure only one file was provided if `multiple` is False
        if not multiple and len(files) != 1:
            logging.warning(
                "Client attempted to upload multiple files when `multiple` was False."
            )
            raise errors.NoFileSelectedError()

        # Return the file info
        if multiple:
            return tuple(files)  # type: ignore
        else:
            return files[0]

    async def _serve_token_validation(
        self, request: fastapi.Request, session_token: str
    ) -> fastapi.Response:
        return fastapi.responses.JSONResponse(
            session_token in self._active_session_tokens
        )

    @contextlib.contextmanager
    def temporarily_disable_new_session_creation(self):
        self._can_create_new_sessions.clear()

        try:
            yield
        finally:
            self._can_create_new_sessions.set()

    async def _serve_websocket(
        self,
        websocket: fastapi.WebSocket,
        sessionToken: str,
    ) -> None:
        """
        Handler for establishing the websocket connection and handling any
        messages.
        """
        # Blah, naming conventions
        session_token = sessionToken
        del sessionToken

        rio._logger.debug(
            f"Received websocket connection with session token `{session_token}`"
        )

        # Wait until we're allowed to accept new websocket connections. (This is
        # temporarily disable while `rio run` is reloading the app.)
        await self._can_create_new_sessions.wait()

        # Accept the socket. I can't figure out how to access the close reason
        # in JS unless the connection was accepted first, so we have to do this
        # even if we are going to immediately close it again.
        await websocket.accept()

        # Look up the session token. If it is valid the session's duration is
        # refreshed so it doesn't expire. If the token is not valid, don't
        # accept the websocket.
        try:
            request = self._latent_session_tokens.pop(session_token)
        except KeyError:
            # Check if this is a reconnect
            try:
                sess = self._active_session_tokens[session_token]
            except KeyError:
                # Inform the client that this session token is invalid
                await websocket.close(
                    3000,  # Custom error code
                    "Invalid session token.",
                )
                return

            # Check if this session still has a functioning websocket
            # connection. Browsers have a "duplicate tab" feature that can
            # create a 2nd tab with the same session token as the original one,
            # and in that case we want to create a new session.
            if sess._transport is not None:
                await websocket.close(
                    3000,  # Custom error code
                    "Invalid session token.",
                )
                return

            # Replace the session's websocket
            sess._transport = transport = FastapiWebsocketTransport(websocket)

            # Make sure the client is in sync with the server by refreshing
            # every single component
            await sess._send_all_components_on_reconnect()

            # Start listening for incoming messages. The session has just
            # received a new websocket connection, so a new task is needed.
            #
            # The previous one was closed when the transport was replaced.
            self._session_serve_tasks[sess] = asyncio.create_task(
                self._serve_session(sess),
                name=f"`Session.serve` for session id `{id(sess)}`",
            )

        else:
            transport = FastapiWebsocketTransport(websocket)

            try:
                sess = await self._create_session_from_websocket(
                    session_token, request, websocket, transport
                )
            except fastapi.WebSocketDisconnect:
                # If the websocket disconnected while we were initializing the
                # session, just close it
                return

            self._active_session_tokens[session_token] = sess
            self._active_tokens_by_session[sess] = session_token

            # Trigger a refresh. This will also send the initial state to
            # the frontend.
            await sess._refresh()

        # Apparently the websocket becomes unusable as soon as this function
        # exits, so we must wait until we no longer need the websocket.
        #
        # When exiting `rio run` with Ctrl+C, this task is cancelled and screams
        # loudly in the console. Suppress that by catching the exception.
        try:
            await transport.closed.wait()
        except asyncio.CancelledError:
            pass

    async def _create_session_from_websocket(
        self,
        session_token: str,
        request: fastapi.Request,
        websocket: fastapi.WebSocket,
        transport: FastapiWebsocketTransport,
    ) -> rio.Session:
        assert request.client is not None, "Why can this happen?"

        # Upon connecting, the client sends an initial message containing
        # information about it. Wait for that, but with a timeout - otherwise
        # evildoers could overload the server with connections that never send
        # anything.
        initial_message_json: Jsonable = await asyncio.wait_for(
            websocket.receive_json(),
            timeout=60,
        )

        initial_message = data_models.InitialClientMessage.from_json(
            initial_message_json  # type: ignore
        )

        try:
            sess = await self.create_session(
                initial_message,
                transport=transport,
                client_ip=request.client.host,
                client_port=request.client.port,
                http_headers=request.headers,
            )
        except routing.NavigationFailed:
            # TODO: Notify the client? Show an error?
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Navigation to initial page `{request.url}` has failed.",
            ) from None

        return sess

    def _after_session_closed(self, session: rio.Session) -> None:
        super()._after_session_closed(session)

        session_token = self._active_tokens_by_session.pop(session)
        del self._active_session_tokens[session_token]



from __future__ import annotations

from typing import *

import rio

from .. import assets, utils
from .abstract_app_server import AbstractAppServer

__all__ = ["TestingServer"]


class TestingServer(AbstractAppServer):
    def weakly_host_asset(self, asset: assets.HostedAsset) -> str:
        # This method is called for font files, even in a test session, so it
        # actually needs to be implemented.
        return asset.secret_id

    def url_for_user_asset(self, relative_asset_path: assets.Path) -> rio.URL:
        raise NotImplementedError

    async def file_chooser(
        self,
        session: rio.Session,
        *,
        file_extensions: Iterable[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | tuple[utils.FileInfo, ...]:
        raise NotImplementedError



import collections
import contextlib
import sys
import tarfile
from pathlib import Path

import revel

from rio.icons.icon_registry import icon_name_to_attr_name

app_name = Path(__file__).stem
app = revel.App(
    nicename=app_name,
    command_name=app_name,
)


@app.command()
def create(archive_file: str, output_file: str | None = None) -> None:
    if output_file is None:
        outfile_ctx = contextlib.nullcontext(sys.stdout)
    else:
        outfile_ctx = open(output_file, "w", encoding="utf8")

    icon_set_name, _, _ = Path(archive_file).name.partition(".")
    icons = collect_icons(archive_file)

    with outfile_ctx as outfile:
        for icon_name in sorted(icons):
            variants = sorted(icons[icon_name])
            for variant in variants:
                attr_name = icon_name_to_attr_name(icon_name, variant=variant)

                if variant:
                    print(
                        f'{attr_name} = "{icon_set_name}/{icon_name}:{variant}"',
                        file=outfile,
                    )
                else:
                    print(
                        f'{attr_name} = "{icon_set_name}/{icon_name}"',
                        file=outfile,
                    )


def collect_icons(archive_file: str) -> dict[str, list[str]]:
    # Map icons to their variants
    icons = collections.defaultdict[str, list[str]](list)

    with tarfile.open(archive_file) as archive:
        for member in archive.getmembers():
            try:
                (icon_set_name, icon_or_variant, *icon) = member.name.split("/")
            except ValueError:
                # The root directory only has 1 path segment, so it throws a
                # ValueError. Skip it.
                continue

            # If `icon` is not empty, then this is an icon with a variant
            if icon:
                icon_name = Path(icon[0]).stem
                icons[icon_name].append(icon_or_variant)
                continue

            # If it's a directory, it's a variant folder. Skip it.
            if member.isdir():
                continue

            # If it's a file, it's an icon with no variant
            icon_name = Path(icon_or_variant).stem
            icons[icon_name].append("")

    return icons


if __name__ == "__main__":
    app.run()



from __future__ import annotations

from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .component import Component
from .fundamental_component import FundamentalComponent

__all__ = [
    "HeadingListItem",
    "SeparatorListItem",
    "SimpleListItem",
    "CustomListItem",
]


@final
class HeadingListItem(FundamentalComponent):
    """
    A list item acting as heading.

    `HeadingListItem`s are an easy way to add structure to your lists. They
    display a short text and style it as heading.

    If your want a more generic list item with additional children, consider
    `SimpleListItem` instead. If that's not enough for your needs you can build
    arbitrarily complex ones using `CustomListItem`.

    ## Attributes

    `text`: The text to display


    ## Examples

    This minimal example will simply display a list item with the text "Click
    me!":

    ```python
    rio.HeadingListItem("Click me!", key="item1")
    ```

    `ListView`s are commonly used to display lists of dynamic length. You can
    easily achieve this by adding the children to a list first, and then
    unpacking that list:

    ```python
    class MyComponent(rio.Component):
        products: list[str] = ["Product 1", "Product 2", "Product 3"]

        def build(self) -> rio.Component:
            # Store all children in an intermediate list
            list_items = []

            for product in self.products:
                list_items.append(
                    rio.HeadingListItem(
                        text=product,
                        key=product,
                    )
                )

            # Then unpack the list to pass the children to the `ListView`
            return rio.ListView(*list_items)
    ```
    """

    text: str


HeadingListItem._unique_id = "HeadingListItem-builtin"


@final
class SeparatorListItem(FundamentalComponent):
    """
    A visual separator between list items.

    `SeparatorListItem`s allow you to group list items visually. They create
    create some empty vertical space between other list items. Normally,
    multiple consecutive list items are also visually grouped, by assigning them
    a card-like background. The `SeparatorListItem` breaks this grouping for
    additional visual separation.
    """


SeparatorListItem._unique_id = "SeparatorListItem-builtin"


@final
class SimpleListItem(Component):
    """
    A simple list item with a header and optional secondary text and children.

    `SimpleListItem`s are a convenient way to create list items, which can
    take care of the most common tasks: Display a text, optional secondary text
    and even additional children (e.g. icons or buttons) to the left and right.
    Most children are optional so you can only add whichever parts you need.


    ## Attributes

    `text`: The text to display.

    `secondary_text`: Additional text to display below the primary text. This
        text may span multiple lines (use `"\\n"` to add a line break).

    `left_child`: A component to display on the left side of the list item.

    `right_child`: A component to display on the right side of the list item.

    `on_press`: Triggered when the list item is pressed.


    ## Examples

    This minimal example will simply display a list item with the text "Click
    me!":

    ```python
    rio.SimpleListItem("Click me!", key="item1")
    ```

    `ListView`s are commonly used to display lists of dynamic length. You can
    easily achieve this by adding the children to a list first, and then
    unpacking that list:

    ```python
    import functools


    class MyComponent(rio.Component):
        products: list[str] = ["Product 1", "Product 2", "Product 3"]

        def on_press_simple_list_item(self, product: str) -> None:
            print(f"Selected {product}")

        def build(self) -> rio.Component:
            # Store all children in an intermediate list
            list_items = []

            for product in self.products:
                list_items.append(
                    rio.SimpleListItem(
                        text=product,
                        key=product,
                        left_child=rio.Icon("material/castle"),
                        # Note the use of `functools.partial` to pass the
                        # product to the event handler.
                        on_press=functools.partial(
                            self.on_press_simple_list_item,
                            product=product,
                        ),
                    )
                )

            # Then unpack the list to pass the children to the `ListView`
            return rio.ListView(*list_items)
    ```
    """

    text: str
    secondary_text: str
    left_child: rio.Component | None
    right_child: rio.Component | None
    on_press: rio.EventHandler[[]]

    def __init__(
        self,
        text: str,
        *,
        key: str | None = None,
        secondary_text: str = "",
        left_child: rio.Component | None = None,
        right_child: rio.Component | None = None,
        on_press: rio.EventHandler[[]] = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            key=key,
        )

        self.text = text
        self.secondary_text = secondary_text
        self.left_child = left_child
        self.right_child = right_child
        self.on_press = on_press

    def build(self) -> rio.Component:
        children = []

        # Left child
        if self.left_child is not None:
            children.append(self.left_child)

        # Main content (text)
        text_children = [
            rio.Text(
                self.text,
                justify="left",
            )
        ]

        if self.secondary_text:
            text_children.append(
                rio.Text(
                    self.secondary_text,
                    wrap=True,
                    style="dim",
                    justify="left",
                )
            )

        children.append(
            rio.Column(
                *text_children,
                spacing=0.5,
                width="grow",
                align_y=0.5,  # In case too much space is allocated
            )
        )

        # Right child
        if self.right_child is not None:
            children.append(self.right_child)

        # Combine everything
        return CustomListItem(
            content=rio.Row(
                *children,
                spacing=1,
                width="grow",
            ),
            on_press=self.on_press,
            key="",
        )


@final
class CustomListItem(FundamentalComponent):
    """
    A list item with custom content.

    Most of the time the `SimpleListItem` will do the job. With
    `CustomListItems` you can build more complex list items. You can add any
    component to the list item. This can be e.g. a `Row`, `Column`, `Text`,
    `Icon`, `Image` or any other component.


    ## Attributes

    `content`: The content to display.

    `on_press`: Triggered when the list item is pressed.

    `key`: A unique key to identify the list item. This is used to avoid
        unintentional reconciliation with other items when the list is
        updated.


    ## Examples

    Instead of using the `SimpleListItem` you can use the `CustomListItem` to create
    a custom list item. This can be useful if you want to add more complex content
    to the list item. You can add any component to the list item.

    ```python
    import functools

    class MyCustomListItemComponent(rio.Component):
        # Create a custom list item
        product: str
        button_text: str

        def build(self) -> rio.Component:
            return rio.Row(
                rio.Text(self.product),
                rio.Spacer(),
                rio.Button(
                    self.button_text,
                    on_press=lambda: print("Hello, world!"),
                ),
            )


    class MyComponent(rio.Component):
        products: list[str] = ["Product 1", "Product 2", "Product 3"]

        def on_press_heading_list_item(self, product: str) -> None:
            print(f"Selected {product}")

        def build(self) -> rio.Component:
            # Store all children in an intermediate list
            list_items = []

            for product in self.products:
                list_items.append(
                    rio.CustomListItem(
                        # Use the `MyCustomListItem` component to create a
                        # custom list item
                        content=MyCustomListItemComponent(
                            product=product, button_text="Click Me!"
                        ),
                        key=product,
                        # Note the use of `functools.partial` to pass the
                        # product to the event handler.
                        on_press=functools.partial(
                            self.on_press_heading_list_item,
                            product=product,
                        ),
                    )
                )

            # Then unpack the list to pass the children to the `ListView`
            return rio.ListView(*list_items)
    ```
    """

    content: rio.Component
    on_press: rio.EventHandler[[]]

    def __init__(
        self,
        content: rio.Component,
        *,
        key: str | None = None,
        on_press: rio.EventHandler[[]] = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            key=key,
        )

        self.content = content
        self.on_press = on_press

    def _custom_serialize(self) -> JsonDoc:
        return {
            "pressable": self.on_press is not None,
        }

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "press", msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        if self.on_press is None:
            return

        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()


CustomListItem._unique_id = "CustomListItem-builtin"



from __future__ import annotations

from typing import *  # type: ignore

import rio

from .. import components as comps


# <component>
class NewsPage(rio.Component):
    """
    A sample page, containing recent news articles about the company.
    """

    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text("Recent News", style="heading1"),
            comps.NewsArticle(
                """
## Disruptive Innovations Inc. Unveils Game-Changing Paradigm Shift at Annual
Synergy Summit

In a move set to redefine industry standards, Disruptive Innovations Inc.
unveiled their latest groundbreaking initiative, the "Quantum Synergy Paradigm,"
at the highly anticipated Annual Synergy Summit yesterday. The event, held at
the epicenter of technological advancement, Silicon Valley, attracted a plethora
of industry thought leaders, gurus, and ninjas. Quantum Synergy Paradigm

The Quantum Synergy Paradigm promises to leverage the power of next-gen AI,
blockchain, nuclear fusion and quantum computing to create a holistic ecosystem
of perpetual growth and exponential ROI. "We are pushing the envelope to unlock
unprecedented value streams for our stakeholders," said CEO Jane Maverick, a
self-proclaimed visionary and industry disruptor. "Our mission is to catalyze a
revolution in how businesses operate, synergizing cross-functional paradigms to
achieve 110% efficiency and 200% customer satisfaction."

Attendees were treated to a spectacular holographic presentation showcasing the
paradigm's potential to integrate seamlessly with existing infrastructures,
reducing friction and maximizing scalability. Early adopters are already hailing
it as a game-changer, predicting a seismic shift in the competitive landscape.

Stay tuned as Disruptive Innovations Inc. continues to blaze trails and shatter
ceilings with their relentless pursuit of excellence and innovation.
                """
            ),
            comps.NewsArticle(
                """
## HyperTech Corp. Acquires Synergistic Solutions LLC in Unprecedented Merger of
Titans

In an unprecedented merger set to revolutionize the tech landscape, HyperTech
Corp. has acquired Synergistic Solutions LLC for a record-breaking $10 billion.
The deal, announced at a press conference brimming with buzzwords and corporate
jargon, promises to deliver unparalleled value to stakeholders and redefine the
boundaries of technological innovation. The Power of Synergy

"This merger is a quantum leap forward in our journey towards market
domination," declared HyperTech CEO John Powerhouse. "By integrating Synergistic
Solutions' disruptive capabilities with our own bleeding-edge technologies, we
are poised to create an unstoppable force of innovation and excellence. This is
not just a merger; it's a synergistic symphony of brilliance."

Synergistic Solutions, renowned for its industry-leading, blue-sky thinking and
holistic approach to problem-solving, will bring a wealth of expertise to
HyperTech's already formidable arsenal. The combined entity is expected to
accelerate the development of hyper-innovative products, leveraging
cross-functional paradigms to achieve unprecedented levels of agility and
operational efficiency.

Industry analysts are already hailing the merger as a masterstroke, predicting
it will catalyze a wave of consolidation and disruption across the tech sector.
The newly formed titan is set to embark on a relentless quest for market
supremacy, driven by a shared vision of exponential growth and unwavering
commitment to customer-centricity.

As HyperTech Corp. and Synergistic Solutions LLC join forces, the world watches
with bated breath, anticipating the dawn of a new era of technological marvels
and corporate synergies.
                """
            ),
            spacing=2,
            width=60,
            margin_bottom=4,
            align_x=0.5,
            align_y=0,
        )


# </component>



import functools
import warnings
from typing import *

from .component_meta import ComponentMeta
from .warnings import *

__all__ = ["deprecated", "parameters_renamed", "_remap_kwargs"]


C = TypeVar("C", bound=Union[Callable, ComponentMeta])
F = TypeVar("F", bound=Callable)


@overload
def deprecated(*, since: str, replacement: Callable): ...


@overload
def deprecated(*, since: str, description: str): ...


def deprecated(
    *,
    since: str,
    description: str | None = None,
    replacement: Callable | None = None,
):
    if replacement is not None:
        description = f"Use {replacement.__qualname__} instead."

    def decorator(callable_: C) -> C:
        callable_.__rio_deprecated_since = since  # type: ignore
        callable_.__rio_deprecated_description = description  # type: ignore
        return callable_

    return decorator


def parameters_renamed(old_names_to_new_names: Mapping[str, str]):
    """
    Class/Function decorator that allows it to be called with the old names of
    the given parameters. For example:

        @parameters_renamed({'foo': 'bar'})
        def example_func(bar: int):
            return bar

        example_func(foo=3)  # Ok at runtime, but static type checker complains
    """

    def decorator(callable_: C) -> C:
        if isinstance(callable_, ComponentMeta):
            callable_._deprecated_parameter_names_.update(
                old_names_to_new_names
            )
            return callable_  # type: ignore (wtf?)

        @functools.wraps(callable_)  # type: ignore (wtf?)
        def wrapper(*args, **kwargs):
            _remap_kwargs(
                callable_.__qualname__, kwargs, old_names_to_new_names
            )
            return callable_(*args, **kwargs)  # type: ignore (wtf?)

        return wrapper  # type: ignore (wtf?)

    return decorator


def parameters_remapped(since: str, **params: Callable[[Any], dict[str, Any]]):
    """
    This is a function decorator that's quite similar to `parameters_renamed`,
    but it allows you to change the type and value(s) of the parameter as well
    as the name.

    The input for the decorator are functions that take the value of the old
    parameter as input and return a dict `{'new_parameter_name': value}`.

    Example: `Theme.from_colors` used to have a `light: bool = True` parameter
    which was changed to `mode: Literal['light', 'dark'] = 'light'`.

        class Theme:
            @parameters_remapped(
                '0.9',
                light=lambda light: {"mode": "light" if light else "dark"},
            )
            def from_colors(..., mode: Literal['light', 'dark'] = 'light'):
                ...

        Theme.from_colors(light=False)  # Equivalent to `mode='dark'`
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for old_name, remap_func in params.items():
                try:
                    old_value = kwargs.pop(old_name)
                except KeyError:
                    pass
                else:
                    [[new_name, new_value]] = remap_func(old_value).items()
                    kwargs[new_name] = new_value

                    warnings.warn(
                        f"The {old_name!r} parameter of rio.{func.__qualname__}"
                        f" is deprecated; please use the {new_name!r} parameter"
                        f" from now on",
                        RioDeprecationWarning,
                    )

            return func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def _remap_kwargs(
    func_name: str,
    kwargs: dict[str, object],
    old_names_to_new_names: Mapping[str, str],
) -> None:
    for old_name, new_name in old_names_to_new_names.items():
        try:
            kwargs[new_name] = kwargs.pop(old_name)
        except KeyError:
            pass
        else:
            warnings.warn(
                f"The {old_name!r} parameter of rio.{func_name} is deprecated; it has been renamed to {new_name!r}",
                RioDeprecationWarning,
            )



import rio


class DocsPage(rio.Component):
    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text(
                "Documentation",
                style="heading2",
                margin=1,
                justify="left",
            ),
            rio.Column(
                rio.Text(
                    "New here? The Rio tutorial can help you get started.",
                    justify="center",
                    wrap=True,
                ),
                rio.Button(
                    "Read the Tutorial",
                    icon="material/school",
                    style="minor",
                    margin=1,
                    on_press=self._open_tutorial,
                ),
                spacing=1,
                height="grow",
                align_y=0.5,
                margin=1,
            ),
        )

    async def _open_tutorial(self) -> None:
        await self.session._evaluate_javascript("""
window.open('https://rio.dev/get-started', '_blank');
""")



import rio


class DeployPage(rio.Component):
    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text(
                "Deploy",
                style="heading2",
                margin=1,
                justify="left",
            ),
            rio.Column(
                rio.Icon(
                    "material/rocket-launch",
                    width=6,
                    height=6,
                    margin_bottom=3,
                    fill=self.session.theme.secondary_color,
                ),
                rio.Text(
                    "One-click deployment is coming soon!",
                    justify="center",
                ),
                rio.Text(
                    "Stay tuned for updates.",
                    justify="center",
                ),
                spacing=1,
                height="grow",
                align_y=0.3,
                margin=1,
            ),
        )



from __future__ import annotations

from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "ThemeContextSwitcher",
]


@final
class ThemeContextSwitcher(FundamentalComponent):
    """
    A container which can switch between theme contexts ("neutral", "warning",
    ...).

    A `ThemeContextSwitcher` is a container which can switch between so-called
    "theme contexts". The idea behind this is that components sometimes need to
    look different based on where they're being used. For example, text inside
    of a `Button` might have a different color than usual.

    Some components, like `Card`s and `Button`s, automatically apply a different
    context to their children. The `ThemeContextSwitcher` gives you the power to
    do this in your own components as well.

    You can find more details on how theming works in the [Theming Quickstart
    Guide](https://rio.dev/docs/howto/theming-guide).


    ## Attributes

    `content`: The currently displayed component.

    `color`: The color of the switcher bar.


    ## Examples

    A minimal example of a `ThemeContextSwitcher`:

    ```python
    rio.ThemeContextSwitcher(
        content=rio.Button("Button"),
        color="secondary",
    )
    ```
    """

    content: rio.Component
    color: rio.ColorSet

    def _custom_serialize(self) -> JsonDoc:
        return {
            "color": self.session.theme._serialize_colorset(self.color),
        }


ThemeContextSwitcher._unique_id = "ThemeContextSwitcher-builtin"



from .fundamental_component import FundamentalComponent

__all__ = [
    "DevToolsConnector",
]


class DevToolsConnector(FundamentalComponent):
    pass


DevToolsConnector._unique_id = "DevToolsConnector-builtin"



from __future__ import annotations

import enum
import functools
import inspect
import json
import types
from typing import *  # type: ignore

import introspection.types
import uniserde
from uniserde import Jsonable, JsonDoc

import rio

from . import color, fills, inspection, maybes, session
from .components import fundamental_component
from .dataclass import class_local_fields
from .self_serializing import SelfSerializing

__all__ = ["serialize_json", "serialize_and_host_component"]


T = TypeVar("T")
Serializer = Callable[["session.Session", T], Jsonable]


FILL_LIKES = {*get_args(fills._FillLike), None, type(None)}


def _float_or_zero(obj: object) -> float:
    try:
        return float(obj)  # type: ignore
    except ValueError:
        return 0


def _serialize_special_types(obj: object) -> Jsonable:
    try:
        func = maybes.TYPE_NORMALIZERS[type(obj)]
    except KeyError:
        raise TypeError(
            f"Can't serialize {obj!r} of type {type(obj)} as JSON"
        ) from None

    return func(obj)  # type: ignore


def _get_margin(*margins: float | None) -> float:
    for margin in margins:
        if margin is not None:
            return margin

    return 0


def serialize_json(data: Jsonable) -> str:
    """
    Like `json.dumps`, but can also serialize numpy types.
    """
    try:
        return json.dumps(data, default=_serialize_special_types)
    except TypeError:
        # Re-initialize the maybes, someone probably imported numpy/pandas after
        # the app was started
        maybes.initialize(force=True)
        return json.dumps(data, default=_serialize_special_types)


def serialize_and_host_component(component: rio.Component) -> JsonDoc:
    """
    Serializes the component, non-recursively. Children are serialized just by
    their `_id`.

    Non-fundamental components must have been built, and their output cached in
    the session.
    """
    result: JsonDoc = {
        "_python_type_": type(component).__name__,
        "_key_": component.key,
        "_rio_internal_": component._rio_internal_,
    }

    # Accessing state properties is pretty slow, so we'll store these in local
    # variables
    width = component.width
    height = component.height

    margin_x = component.margin_x
    margin_y = component.margin_y
    margin = component.margin

    # Add layout properties, in a more succinct way than sending them separately
    result["_margin_"] = (
        _get_margin(component.margin_left, margin_x, margin),
        _get_margin(component.margin_top, margin_y, margin),
        _get_margin(component.margin_right, margin_x, margin),
        _get_margin(component.margin_bottom, margin_y, margin),
    )
    # The width/height can be floats or strings, but they could also be numpy
    # floats or numpy strings. We could check `isinstance(width,
    # maybes.STR_TYPES)`, but that would give incorrect output if numpy was
    # imported after the app was launched. I think the easiest solution is to
    # simply try converting it to a float, and if that fails, default to 0.
    # (Although that has the side effect of treating strings like `"1.5"` as
    # numbers.)
    result["_size_"] = (_float_or_zero(width), _float_or_zero(height))
    result["_align_"] = (
        component.align_x,
        component.align_y,
    )
    result["_grow_"] = (
        width == "grow",
        height == "grow",
    )

    # If it's a fundamental component, serialize its state because JS needs it.
    # For non-fundamental components, there's no reason to send the state to
    # the frontend.
    if isinstance(component, fundamental_component.FundamentalComponent):
        sess = component.session

        for name, serializer in get_attribute_serializers(
            type(component)
        ).items():
            result[name] = serializer(sess, getattr(component, name))

        # Encode any internal additional state. Doing it this late allows the custom
        # serialization to overwrite automatically generated values.
        result["_type_"] = component._unique_id
        result.update(component._custom_serialize())

    else:
        # Take care to add underscores to any properties here, as the
        # user-defined state is also added and could clash
        result["_type_"] = "Placeholder"
        result["_child_"] = component.session._weak_component_data_by_component[
            component
        ].build_result._id

    return result


@functools.lru_cache(maxsize=None)
def get_attribute_serializers(
    cls: Type[rio.Component],
) -> Mapping[str, Serializer]:
    """
    Returns a dictionary of attribute names to their types that should be
    serialized for the given component class.
    """
    serializers: dict[str, Serializer] = {}

    for base_cls in reversed(cls.__bases__):
        if issubclass(base_cls, rio.Component):
            serializers.update(get_attribute_serializers(base_cls))

    annotations = inspection.get_local_annotations(cls)

    for attr_name, field in class_local_fields(cls).items():
        if not field.serialize:
            assert (
                attr_name not in serializers
            ), f"A base class wants to serialize {attr_name}, but {cls} doesn't"
            continue

        serializer = _get_serializer_for_annotation(annotations[attr_name])
        if serializer is None:
            continue

        serializers[attr_name] = serializer

    return serializers


def _serialize_basic_json_value(
    sess: session.Session, value: Jsonable
) -> Jsonable:
    return value


def _serialize_self_serializing(
    sess: session.Session, obj: SelfSerializing
) -> Jsonable:
    return obj._serialize(sess)


def _serialize_child_component(
    sess: session.Session, component: rio.Component
) -> Jsonable:
    return component._id


def _serialize_list(
    sess: session.Session, list_: list[T], item_serializer: Serializer[T]
) -> Jsonable:
    return [item_serializer(sess, item) for item in list_]


def _serialize_enum(
    sess: session.Session, value: object, as_type: Type[enum.Enum]
) -> Jsonable:
    return uniserde.as_json(value, as_type=as_type)


def _serialize_colorset(
    sess: session.Session, colorset: color.ColorSet
) -> Jsonable:
    return sess.theme._serialize_colorset(colorset)


def _serialize_fill_like(
    sess: session.Session, fill: fills._FillLike | None
) -> Jsonable:
    return sess._serialize_fill(fill)


def _serialize_optional(
    sess: session.Session, value: T | None, serializer: Serializer[T]
) -> Jsonable:
    if value is None:
        return None

    return serializer(sess, value)


def _get_serializer_for_annotation(
    annotation: introspection.types.TypeAnnotation,
) -> Serializer | None:
    """
    Which values are serialized for state depends on the annotated datatypes.
    There is no point in sending fancy values over to the client which it can't
    interpret.

    This function looks at the annotation and returns a suitable serialization
    function, or `None` if this attribute shouldn't be serialized.
    """
    # Basic JSON values
    if annotation in (int, float, str, bool, None):
        return _serialize_basic_json_value

    origin = get_origin(annotation)
    args = get_args(annotation)

    # Python 3.10 crashes if you try `issubclass(list[str], SelfSerializing)`,
    # so we must make absolutely sure the annotation isn't a generic type
    if inspect.isclass(annotation) and not args:
        # Self-Serializing
        if issubclass(annotation, SelfSerializing):
            return _serialize_self_serializing

        # Components
        if issubclass(annotation, rio.Component):
            return _serialize_child_component

        # Enums
        if issubclass(annotation, enum.Enum):
            return functools.partial(_serialize_enum, as_type=annotation)

    # Sequences of serializable values
    if origin is list:
        item_serializer = _get_serializer_for_annotation(args[0])
        if item_serializer is None:
            return None
        return functools.partial(
            _serialize_list, item_serializer=item_serializer
        )

    # Literal
    if origin is Literal:
        return _serialize_basic_json_value

    if origin in (Union, types.UnionType):
        # ColorSet
        if set(args) == color._color_set_args:
            return _serialize_colorset

        # Optional
        if len(args) == 2 and type(None) in args:
            type_ = next(type_ for type_ in args if type_ is not type(None))
            serializer = _get_serializer_for_annotation(type_)
            if serializer is None:
                return None
            return functools.partial(_serialize_optional, serializer=serializer)

        # Fills
        #
        # Note: FILL_LIKES includes `None` and `Color`. We don't want a `None |
        # Color` to be serialized as a Fill, so this code must be below the
        # `Optional` check.
        if set(args) <= FILL_LIKES:
            return _serialize_fill_like

    return None



# <additional-imports>
import openai  # type: ignore (hidden from user)

import rio

# </additional-imports>


# <additional-code>
OPENAI_API_KEY = "<placeholder>"  # Replace this with your OpenAI API key


# Instruct the developer to replace the placeholder key if they haven't done so
# yet. Feel free to delete this code if you've already replaced the key.

if OPENAI_API_KEY == "<placeholder>":
    message = """
This template requires an OpenAI API key to work

You can get your API key from [OpenAI's website](https://platform.openai.com/api-keys
Make sure to enter your key into the `__init__.py` file before trying to run the project
""".strip()

    print(message)
    raise RuntimeError(message)


def on_app_start(app: rio.App):
    # Create the OpenAI client and attach it to the app
    app.default_attachments.append(openai.AsyncOpenAI(api_key=OPENAI_API_KEY))


# </additional-code>
